"""Algorithms for EnergyFlow."""
from __future__ import absolute_import

from . import einsumfunc
from . import integer_partitions
from . import ve

from .einsumfunc import *
from .integer_partitions import *
from .ve import *

__all__ = einsumfunc.__all__ + integer_partitions.__all__ + ve.__all__



from __future__ import absolute_import, division, print_function

from abc import ABCMeta, abstractmethod, abstractproperty

from keras.optimizers import Adam
from six import with_metaclass

__all__ = [
    'ArchBase',
    'NNBase'
]

###############################################################################
# ArchBase
###############################################################################
class ArchBase(with_metaclass(ABCMeta, object)):

    """Base class for all architectures contained in EnergyFlow. The mechanism of
    specifying hyperparameters for all architectures is described here. Methods
    common to all architectures are documented here. Note that this class cannot
    be instantiated directly as it is an abstract base class.
    """

    # ArchBase(*args, **kwargs)
    def __init__(self, *args, **kwargs):
        """Accepts arbitrary arguments. Positional arguments (if present) are
        dictionaries of hyperparameters, keyword arguments (if present) are 
        hyperparameters directly. Keyword hyperparameters take precedence over
        positional hyperparameter dictionaries.

        **Arguments**

        - ***args** : arbitrary positional arguments
            - Each argument is a dictionary containing hyperparameter (name, value)
            pairs.
        - ***kwargs** : arbitrary keyword arguments
            - Hyperparameters as keyword arguments. Takes precedence over the 
            positional arguments.
        """
        
        # store all options
        self.hps = {}
        for d in args:
            self.hps.update(d)
        self.hps.update(kwargs)

        # process hyperparameters
        self.process_hps()

        # construct model
        self.construct_model()

    @abstractmethod
    def process_hps(self):
        pass

    @abstractmethod
    def construct_model(self):
        pass

    # fit(X_train, Y_train, **kwargs)
    @abstractmethod
    def fit(self):
        """Train the model by fitting the provided training dataset and labels.
        Transparently calls the `fit()` method of the underlying model.

        **Arguments**

        - **X_train** : _numpy.ndarray_
            - The training dataset as an array of features for each sample.
        - **Y_train** : _numpy.ndarray_
            - The labels for the training dataset. May need to be one-hot encoded
            depending on the requirements of the underlying model (typically Keras
            models will use one-hot encoding whereas the linear model does not.)
        - **kwargs** : _dict_
            - Keyword arguments passed on to the Keras method of the same name.
            See the [Keras model docs](https://keras.io/models/model/#fit) for
            details on available parameters.

        **Returns**

        - Whatever the underlying model's `fit()` returns.
        """

        pass

    # predict(X_test, **kwargs)
    @abstractmethod
    def predict(self):
        """Evaluate the model on a dataset. 

        **Arguments**

        - **X_test** : _numpy.ndarray_
            - The dataset to evaluate the model on.
        - **kwargs** : _dict_
            - Keyword arguments passed on to the Keras method of the same name.
            See the [Keras model docs](https://keras.io/models/model/#fit) for
            details on available parameters.

        **Returns**

        - _numpy.ndarray_
            - The value of the model on the input dataset.
        """

        pass

    @abstractproperty
    def model(self):
        """The underlying model held by this architecture."""

        pass

###############################################################################
# NNBase
###############################################################################
class NNBase(ArchBase):        

    def process_hps(self):
        """**Default NN Hyperparameters**

        Common hyperparameters that apply to all architectures except 
        for [`LinearClassifier`](#linearclassifier).

        - **loss**=`'categorical_crossentropy'` : _str_
            - The loss function to use for the model. See the [Keras
            loss function docs](https://keras.io/losses/) for available
            loss functions.
        - **lr**=`0.001` : _float_
            - The learning rate for the model.
        - **opt**=`Adam` : Keras optimizer
            - A [Keras optimizer](https://keras.io/optimizers/).
        - **output_dim**=`2` : _int_
            - The output dimension of the model.
        - **output_act**=`'softmax'` : _str_
            - Activation function to apply to the output.
        - **metrics**=`['accuracy']` : _list_ of _str_
            - The [Keras metrics](https://keras.io/metrics/) to apply
            to the model.
        - **compile**=`True` : _bool_
            - Whether the model should be compiled or not.
        - **summary**=`True` : _bool_
            - Whether a summary should be printed or not.
        """

        # optimization
        self.loss = self.hps.get('loss', 'categorical_crossentropy')
        self.lr = self.hps.get('lr', 0.001)
        self.opt = self.hps.get('opt', Adam)

        # output
        self.output_dim = self.hps.get('output_dim', 2)
        self.output_act = self.hps.get('output_act', 'softmax')

        # metrics
        self.metrics = self.hps.get('metrics', ['accuracy'])

        # flags
        self.compile = self.hps.get('compile', True)
        self.summary = self.hps.get('summary', True)

    def compile_model(self):

        # compile model if specified
        if self.compile: 
            self.model.compile(loss=self.loss, optimizer=self.opt(lr=self.lr), metrics=self.metrics)

            # print summary
            if self.summary: 
                self.model.summary()

    def fit(self, *args, **kwargs):
        return self.model.fit(*args, **kwargs)

    def predict(self, *args, **kwargs):
        return self.model.predict(*args, **kwargs)

    @property
    def model(self):
        return self._model



from __future__ import absolute_import, division, print_function

from abc import abstractmethod

from keras import backend as K
from keras.layers import Dense, Dot, Dropout, Input, Lambda, TimeDistributed, Masking
from keras.models import Model
from keras.regularizers import l2
import numpy as np

from energyflow.archs.archbase import NNBase
from energyflow.utils import iter_or_rep

__all__ = ['EFN', 'PFN']

###############################################################################
# SymmetricPerParticleNN - Base class for EFN-like models
###############################################################################
class SymmetricPerParticleNN(NNBase):

    # EFN(*args, **kwargs)
    def process_hps(self):
        """See [`ArchBase`](#archbase) for how to pass in hyperparameters.

        **Required EFN Hyperparameters**

        - **input_dim** : _int_
            - The number of features for each particle.
        - **ppm_sizes** : {_tuple_, _list_} of _int_
            - The sizes of the dense layers in the per-particle frontend
            module $\\Phi$. The last element will be the number of latent 
            observables that the model defines.
        - **dense_sizes** : {_tuple_, _list_} of _int_
            - The sizes of the dense layers in the backend module $F$.

        **Default EFN Hyperparameters**

        - **ppm_acts**=`'relu'` : {_tuple_, _list_} of _str_
            - Activation functions(s) for the dense layers in the 
            per-particle frontend module $\\Phi$. A single string will apply 
            the same activation to all layers. See the [Keras activations 
            docs](https://keras.io/activations/) for more detail.
        - **dense_acts**=`'relu'` : {_tuple_, _list_} of _str_
            - Activation functions(s) for the dense layers in the 
            backend module $F$. A single string will apply  the same activation 
            to all layers.
        - **ppm_k_inits**=`'he_uniform'` : {_tuple_, _list_} of _str_
            - Kernel initializers for the dense layers in the per-particle
            frontend module $\\Phi$. A single string will apply the same initializer 
            to all layers. See the [Keras initializer docs](https://
            keras.io/initializers/) for more detail.
        - **dense_k_inits**=`'he_uniform'` : {_tuple_, _list_} of _str_
            - Kernel initializers for the dense layers in the backend 
            module $F$. A single string will apply the same initializer 
            to all layers.
        - **latent_dropout**=`0` : _float_
            - Dropout rates for the summation layer that defines the
            value of the latent observables on the inputs. See the [Keras
            Dropout layer](https://keras.io/layers/core/#dropout) for more 
            detail.
        - **dense_dropouts**=`0` : {_tuple_, _list_} of _float_
            - Dropout rates for the dense layers in the backend module $F$. 
            A single float will apply the same dropout rate to all dense layers.
        - **mask_val**=`0` : _float_
            - The value for which particles with all features set equal to
            this value will be ignored. See the [Keras Masking layer](https://
            keras.io/layers/core/#masking) for more detail.
        """

        # process generic NN hps
        super(SymmetricPerParticleNN, self).process_hps()

        # required hyperparameters
        self.input_dim = self.hps['input_dim']
        self.ppm_sizes = self.hps['ppm_sizes']
        self.dense_sizes = self.hps['dense_sizes']

        # activations
        self.ppm_acts = iter_or_rep(self.hps.get('ppm_acts', 'relu'))
        self.dense_acts = iter_or_rep(self.hps.get('dense_acts', 'relu'))

        # initializations
        self.ppm_k_inits = iter_or_rep(self.hps.get('ppm_k_inits', 'he_uniform'))
        self.dense_k_inits = iter_or_rep(self.hps.get('dense_k_inits', 'he_uniform'))

        # regularizations
        #self.ppm_dropouts = iter_or_rep(self.hps.get('ppm_dropouts', 0))
        self.latent_dropout = self.hps.get('latent_dropout', 0)
        self.dense_dropouts = iter_or_rep(self.hps.get('dense_dropouts', 0))

        # masking
        self.mask_val = self.hps.get('mask_val', 0.)

    @abstractmethod
    def construct_input_layers(self):
        pass

    def construct_per_particle_module(self):

        # a list of the per-particle layers, starting with the masking layer operating on input 0
        self.ppm_layers = [Masking(mask_value=self.mask_val, name='mask_0')(self.input_layers[-1])]

        # iterate over specified layers
        for i,(s, act, k_init) in enumerate(zip(self.ppm_sizes, self.ppm_acts, self.ppm_k_inits)):

            # define a dense layer that will be applied through time distributed
            d_layer = Dense(s, activation=act, kernel_initializer=k_init)

            # append time distributed layer to list of ppm layers
            self.ppm_layers.append(TimeDistributed(d_layer, name='tdist_'+str(i))(self.ppm_layers[-1]))

    @abstractmethod
    def construct_latent_layer(self):
        pass

    def construct_backend_module(self):
        
        # a list of backend layers
        self.backend_layers = [self.latent_layer]

        # iterate over specified backend layers
        z = zip(self.dense_sizes, self.dense_acts, self.dense_k_inits, self.dense_dropouts)
        for i,(s, act, k_init, dropout) in enumerate(z):

            # a new dense layer
            new_layer = Dense(s, activation=act, kernel_initializer=k_init, name='dense_'+str(i))

            # apply dropout if specified 
            if dropout > 0:
                new_layer = Dropout(dropout, name='dropout_'+str(i))(new_layer)

            # apply new layer to previous and append to list
            self.backend_layers.append(new_layer(self.backend_layers[-1]))

    def construct_model(self):

        # construct earlier parts of the model
        self.construct_input_layers()
        self.construct_per_particle_module()
        self.construct_latent_layer()
        self.construct_backend_module()

        # output layer, applied to the last backend layer
        output_layer = Dense(self.output_dim, activation=self.output_act, 
                                              name='output')(self.backend_layers[-1])

        # construct a new model
        self._model = Model(inputs=self.input_layers, outputs=output_layer)

        # compile model
        self.compile_model()

    @property
    def input_layers(self):
        return self._input_layers

    @property
    def latent_layer(self):
        return self._latent_layer


###############################################################################
# EFN - Energy flow network class
###############################################################################

class EFN(SymmetricPerParticleNN):

    """Energy Flow Network (EFN) architecture."""

    def construct_input_layers(self):

        zs_input = Input(batch_shape=(None, None), name='zs_input')
        phats_input = Input(batch_shape=(None, None, self.input_dim), name='phats_input')
        self._input_layers = [zs_input, phats_input]

    def construct_latent_layer(self):

        self._latent_layer = Dot(0, name='dot')([self.input_layers[0], self.ppm_layers[-1]])

        if self.latent_dropout > 0:
            self._latent_layer = Dropout(self.latent_dropout, name='latent_dropout')(self._latent_layer)

    # eval_filters(patch, n=100, prune=True)
    def eval_filters(self, patch, n=100, prune=True):
        """Evaluates the latent space filters of this model on a patch of the 
        two-dimensional geometric input space.

        **Arguments**

        - **patch** : {_tuple_, _list_} of _float_
            - Specifies the patch of the geometric input space to be evaluated.
            A list of length 4 is interpretted as `[xmin, ymin, xmax, ymax]`.
            Passing a single float `R` is equivalent to `[-R,-R,R,R]`.
        - **n** : {_tuple_, _list_} of _int_
            - The number of grid points on which to evaluate the filters. A list 
            of length 2 is interpretted as `[nx, ny]` where `nx` is the number of
            points along the x (or first) dimension and `ny` is the number of points
            along the y (or second) dimension.
        - **prune** : _bool_
            - Whether to remove filters that are all zero (which happens sometimes
            due to dying ReLUs).

        **Returns**

        - (_numpy.ndarray_, _numpy.ndarray_, _numpy.ndarray_)
            - Returns three arrays, `(X, Y, Z)`, where `X` and `Y` have shape `(nx, ny)` 
            and are arrays of the values of the geometric inputs in the specified patch.
            `Z` has shape `(num_filters, nx, ny)` and is the value of the different
            filters at each point.
        """

        # determine patch of xy space to evaluate filters on
        if isinstance(patch, (float, int)):
            if patch > 0:
                xmin, ymin, xmax, ymax = -patch, -patch, patch, patch
            else:
                ValueError('patch must be positive when passing as a single number.')
        else:
            xmin, ymin, xmax, ymax = patch

        # determine number of pixels in each dimension
        if isinstance(n, int):
            nx = ny = n
        else:
            nx, ny = n

        # construct grid of inputs
        xs, ys = np.linspace(xmin, xmax, nx), np.linspace(ymin, ymax, ny)
        X, Y = np.meshgrid(xs, ys, indexing='ij')
        XY = np.asarray([X, Y]).reshape((2, nx*ny)).T

        # construct function 
        kf = K.function([self.input_layers[1]], [self.ppm_layers[-1]])

        # evaluate function
        Z = kf([[XY]])[0][0].reshape(nx, ny, self.ppm_sizes[-1]).transpose((2,0,1))

        # prune filters that are off
        if prune:
            return X, Y, Z[[not (z == 0).all() for z in Z]]
        
        return X, Y, Z


###############################################################################
# PFN - Particle flow network class
###############################################################################

class PFN(SymmetricPerParticleNN):

    """Particle Flow Network (PFN) architecture. Accepts the same 
    hyperparameters as the [`EFN`](#EFN)."""

    # PFN(*args, **kwargs)
    def construct_input_layers(self):
        """""" # need this for autogen docs

        self._input_layers = [Input(batch_shape=(None, None, self.input_dim), name='input')]

    def construct_latent_layer(self):

        self._latent_layer = Lambda(lambda x: K.sum(x, axis=1), name='sum')(self.ppm_layers[-1])

        if self.latent_dropout > 0:
            self._latent_layer = Dropout(self.latent_dropout, name='latent_dropout')(self._latent_layer)



"""### Random Events

Functions to generate random sets of four-vectors. Includes an implementation
of the [RAMBO](https://doi.org/10.1016/0010-4655(86)90119-0) algorithm for
sampling uniform M-body massless phase space. Also includes other functions for
various random, non-center of momentum, and non-uniform sampling.
"""
from __future__ import absolute_import

import os

import numpy as np

__all__ = [
    'gen_random_events',
    'gen_random_events_mcom',
    'gen_massless_phase_space'
]

def gen_random_events(nevents, nparticles, dim=4, mass=0):
    """Generate random events with a given number of particles of a given mass
    in a given spacetime dimension. The energy-momentum vectors have spatial
    components drawn randomly from [-1,+1]. These events are not guaranteed
    to uniformly sample phase space.

    **Arguments**

    - **nevents** : _int_
        - Number of events to generate.
    - **nparticles** : _int_
        - Number of particles in each event.
    - **dim** : _int_
        - Number of spacetime dimensions.
    - **mass** : _float_
        - Mass of the particles to generate.

    **Returns**

    - _numpy.ndarray_
        - An (`nevents`, `nparticles`, `dim`) array of events, each with `nparticles` particles
        with mass given by `mass`.
    """

    spatial_ps = 2*np.random.rand(nevents, nparticles, dim-1) - 1
    energies = np.sqrt(mass**2 + np.sum(spatial_ps**2, axis=-1))
    events = np.concatenate((energies[:,:,np.newaxis], spatial_ps), axis=-1) 
    return np.squeeze(events)

def gen_random_events_mcom(nevents, nparticles, dim=4):
    """Generate random events with a given number of massless particles
    in a given spacetime dimension. The total energy and momentum are made to sum to zero
    by making about half of the particles incoming. These events are not guaranteed
    to uniformly sample phase space.

    **Arguments**

    - **nevents** : _int_
        - Number of events to generate.
    - **nparticles** : _int_
        - Number of particles in each event.
    - **dim** : _int_
        - Number of spacetime dimensions.

    **Returns**

    - _numpy.ndarray_
        - An (`nevents`, `nparticles`, `dim`) array of events, each with `nparticles` massless
        particles whose total energy and momentum are all zero.
    """

    events_1_sp = 2*np.random.rand(nevents, int(np.ceil(nparticles/2)-1), dim-1) - 1
    events_2_sp = 2*np.random.rand(nevents, int(np.floor(nparticles/2)-1), dim-1) - 1
    
    events_1_sp_com = np.concatenate((events_1_sp, -np.sum(events_1_sp, axis=1)[:,np.newaxis]), axis=1)
    events_2_sp_com = np.concatenate((events_2_sp, -np.sum(events_2_sp, axis=1)[:,np.newaxis]), axis=1)

    events_1_tup = (np.sqrt(np.sum(events_1_sp_com**2, axis=-1))[:,:,np.newaxis], events_1_sp_com)
    events_2_tup = (np.sqrt(np.sum(events_2_sp_com**2, axis=-1))[:,:,np.newaxis], events_2_sp_com)
    events_1 = np.concatenate(events_1_tup, axis=-1)
    events_2 = np.concatenate(events_2_tup, axis=-1)

    events_1_tot, events_2_tot = np.sum(events_1, axis=1), np.sum(events_2, axis=1)
    factors = events_1_tot[:,0]/events_2_tot[:,0]

    return np.concatenate((events_1, -events_2*factors[:,np.newaxis,np.newaxis]), axis=1)

def gen_massless_phase_space(nevents, nparticles, energy=1):
    """Implementation of the [RAMBO](https://doi.org/10.1016/0010-4655(86)90119-0) algorithm
    for uniformly sampling massless M-body phase space for any center of mass energies.
    
    **Arguments**

    - **nevents** : _int_
        - Number of events to generate.
    - **nparticles** : _int_
        - Number of particles in each event.
    - **energy** : _float_
        - Total center of mass energy of each event.

    **Returns**

    - _numpy.ndarray_
        - An (`nevents`, `nparticles`, 4) array of events, each with `nparticles` massless
        particles and center of mass energy equal to `energy`.
    """
    
    # qs: to be massless four-momenta uniformly sampled in angle
    qs = np.empty((nevents, nparticles, 4))
    
    # ps: to be the uniformly sampled n-body four-momenta s.t. sum_i p_i = (energy, 0)
    ps = np.empty((nevents, nparticles, 4))
    
    # randomly sample from the qs as stated in the RAMBO paper
    r = np.random.random((4, nevents, nparticles))

    c = 2*r[0] - 1
    phi = 2*np.pi*r[1]
        
    qs[:,:,0] = -np.log(r[2]*r[3])
    tmp = qs[:,:,0]*np.sqrt(1 - c**2)
    qs[:,:,1] = tmp*np.cos(phi)
    qs[:,:,2] = tmp*np.sin(phi)
    qs[:,:,3] = qs[:,:,0]*c
        
    # define the following quantities to rescale the qs to the ps
    Qs = np.sum(qs, axis=1)
    Ms = np.sqrt(np.abs(Qs[:,0]**2 - Qs[:,1]**2 - Qs[:,2]**2 - Qs[:,3]**2))
    gammas = Qs[:,0]/Ms
    As = 1/(1 + gammas)[:,np.newaxis,np.newaxis]
    bs = (-Qs[:,1:]/Ms[:,np.newaxis])[:,np.newaxis]
    xs = (energy/Ms)[:,np.newaxis]

    bdotq = np.sum(bs*qs[:,:,1:], axis=-1)

    ps[:,:,0] = xs*(gammas[:,np.newaxis]*qs[:,:,0] + bdotq)
    ps[:,:,1:] = xs[:,np.newaxis]*(qs[:,:,1:] + 
                                              bs*qs[:,:,0,np.newaxis] + 
                                              As*bdotq[:,:,np.newaxis]*bs)
    return np.squeeze(ps)



"""### Quark and Gluon Jets

A dataset consisting of up to 2 million total quark and gluon jets generated with PYTHIA 8.226.
To avoid downloading unnecessary samples, the dataset is contained in twenty
files with 100k jets each, and only the required files are downloaded.
These samples are used in [1810.05165](https://arxiv.org/abs/1810.05165).
Splitting the data into 1.6M/200k/200k train/validation/test sets is recommended for standardized comparisons.

The dataset `qg_jets` consists of two components:

* `X` : a three-dimensional numpy array of the jets with shape `(num_data, max_num_particles, 4)`
* `y` : a numpy array of quark/gluon jet labels 
(quark=`1` and gluon=`0`).

The jets are padded with zero-particles in order to make a contiguous 
array. The particles are given as `(pt,y,phi,pid)` values, where `pid` is the particle's 
[PDG id](http://pdg.lbl.gov/2007/reviews/montecarlorpp.pdf).


The samples are $Z(\\to\\nu\\bar\\nu)+g$ and $Z(\\to\\nu\\bar\\nu)+(u,d,s)$
 events generated with PYTHIA for $pp$ collisions at $\sqrt{s}=14$ TeV using the `WeakBosonAndParton:qqbar2gmZg` 
 and `WeakBosonAndParton:qg2gmZq` processes, ignoring the photon contribution and requiring
 the $Z$ to decay invisibly to neutrinos.
Hadronization and multiple parton interactions (i.e. underlying event) are turned on
 and the default tunings and shower parameters are used.
Final state non-neutrino particles are clustered into $R=0.4$ anti-$k_T$ jets
 using FASTJET 3.3.0.
Jets with transverse momentum $p_T\in[500,550]$ GeV and rapidity $|y|<2.0$ are kept.
Particles are ensured to have $\phi$ values within $\pi$ of the jet (i.e. no $\phi$-periodicity issues).
No detector simulation is performed.
"""
from __future__ import absolute_import

import warnings

import numpy as np

from energyflow.utils.data_utils import _get_file

__all__ = ['load']

QG_jets_urls = [
    'https://www.dropbox.com/s/fclsl7pukcpobsb/QG_jets.npz?dl=1',
    'https://www.dropbox.com/s/ztzd1a6lkmgovuy/QG_jets_1.npz?dl=1',
    'https://www.dropbox.com/s/jzgc9e786tbk1m5/QG_jets_2.npz?dl=1',
    'https://www.dropbox.com/s/tiwz2ck3wnzvlcr/QG_jets_3.npz?dl=1',
    'https://www.dropbox.com/s/3miwek1n0brbd2i/QG_jets_4.npz?dl=1',
    'https://www.dropbox.com/s/tsq80wc6ngen9kn/QG_jets_5.npz?dl=1',
    'https://www.dropbox.com/s/5oba2h15ufa57ie/QG_jets_6.npz?dl=1',
    'https://www.dropbox.com/s/npl6b2rts82r1ya/QG_jets_7.npz?dl=1',
    'https://www.dropbox.com/s/7pldxfqdb4n0kaw/QG_jets_8.npz?dl=1',
    'https://www.dropbox.com/s/isw4clv7n370nfb/QG_jets_9.npz?dl=1',
    'https://www.dropbox.com/s/prw7myb889v2y12/QG_jets_10.npz?dl=1',
    'https://www.dropbox.com/s/10r4ydro3e6nsmc/QG_jets_11.npz?dl=1',
    'https://www.dropbox.com/s/42p10sv9jedmtn0/QG_jets_12.npz?dl=1',
    'https://www.dropbox.com/s/crqdeg4arjti7cy/QG_jets_13.npz?dl=1',
    'https://www.dropbox.com/s/1e7ss2quxhkbhwy/QG_jets_14.npz?dl=1',
    'https://www.dropbox.com/s/psje9feje43buc7/QG_jets_15.npz?dl=1',
    'https://www.dropbox.com/s/8qw5bcswgrr9fl1/QG_jets_16.npz?dl=1',
    'https://www.dropbox.com/s/gcdp98bgupfk05x/QG_jets_17.npz?dl=1',
    'https://www.dropbox.com/s/jvgt17z1ufxz1ly/QG_jets_18.npz?dl=1',
    'https://www.dropbox.com/s/gbbfvy2e0slmm8v/QG_jets_19.npz?dl=1',
]

QG_jets_hashes = [
    '3f27a02eab06e8b83ccc9d25638021e6e24c9361341730961f9d560dee12c257',
    '648e49cd59b5353e0064e7b1a3388d9c2f4a454d3ca67afaa8d0344c836ecb35',
    '09f7b16fa7edb312c0f652bb8504de45f082c4193df65204d693155017272fe9',
    '7dc9a50bb38e9f6fc1f11db18f9bd04f72823c944851746b848dee0bba808537',
    '3e6217aad8e0502f5ce3b6371c61396dfc48a6cf4f26ee377cc7b991b1d2b543',
    'b5b7d742b2599bcbe1d7a639895bca64c28da513dc3620b0e5bbb5801f8c88fd',
    '7d31bc48c15983401e0dbe8fd5ee938c3809d9ee3c909f4adab6daf8b73c14f1',
    'cec0d7b2afa9d955543c597f9b7f3b3767812a68b2401ec870caf3a2ceb98401',
    'e984620f57abe06fc5d0b063f9f84ba54bd3e8c295d2b2419a7b1c6175079ed4',
    '6e3b69196995d6eb3b8e7af874e2b9f93d904624f7a7a73b8ff39f151e3bd189',
    'fa3d386f230b806058ff17e5bd77326ff4bf01d72aa5eb3325c1df2a8825927c',
    'acd49ab7bea8f72ecf699a9a898bccacc8730474259d68406656a5a43d407fb0',
    '2edd55b8bc30c686a0637855e1ba068586eb97041e8114d5540d96db2a7a2e17',
    '7276a8a0e573f9795a47f9d5addc10d2af903c2a0ffa5c848a720ccae93daa90',
    '2068ecfa912e94cd3ce7273b7c77af0bbd5ec57940997e7483b56f03434a6869',
    '41a732ce6321dd593214225b03fb87329607ccae768c705e3896ffecc28bfcca',
    '9d68caeb18f3ccf127b9032f52e63ee011c4381293a3a503f894e5c0741ae215',
    '086053ca611bb04d97fa0b6509b4ffb6955421b067c7b277498f0e5188879331',
    'cdc595f5fedef7db9411a9f93f2786f110073b4d17a523700f625846588b1e44',
    'd07781139320ae134ce4824bc0cefa43fd5003cd97cdf3aed90d4fb12fad8a1d',
]

num_per_file = 100000
max_num_files = len(QG_jets_urls)

def load(num_data=100000, pad=True, cache_dir=None):
    """Loads samples from the dataset (which in total is contained in twenty files). 
    Any file that is needed that has not been cached will be automatically downloaded.
    Downloading a file causes it to be cached for later use. Basic checksums are
    performed.

    **Arguments**

    - **num_data** : _int_
        - The number of events to return. A value of `-1` means read in all events.
    - **pad** : _bool_
        - Whether to pad the events with zeros to make them the same length.
    - **cache_dir** : _str_
        - The directory where to store/look for the file.

    **Returns**

    - _3-d numpy.ndarray_, _1-d numpy.ndarray_
        - The `X` and `y` components of the dataset as specified above.
    """

    num_files = int(np.ceil(num_data/num_per_file)) if num_data > -1 else max_num_files

    # handle request for too much data
    if num_files > max_num_files:
        warnings.warn('More data requested than available. Providing the full dataset.')
        num_files = max_num_files
        num_data = -1

    Xs, ys = [], []
    for i in range(num_files):

        # preserve old first file
        filename = 'QG_jets_{}.npz'.format(i) if i > 0 else 'QG_jets.npz'

        f = np.load(_get_file(filename, url=QG_jets_urls[i], 
                                        file_hash=QG_jets_hashes[i], 
                                        cache_dir=cache_dir))
        Xs.append(f['X'])
        ys.append(f['y'])
        f.close()

    if pad:
        max_len_axis1 = max([X.shape[1] for X in Xs])
        X = np.vstack([_pad_events_axis1(x, max_len_axis1) for x in Xs])
    else:
        X = np.asarray([x[x[:,0]>0] for X in Xs for x in X])

    y = np.concatenate(ys)

    if num_data > -1:
        X, y = X[:num_data], y[:num_data]

    return X, y


def _pad_events_axis1(events, axis1_shape):
    """Pads the first axis of the NumPy array `events` with zero subarrays
    such that the first dimension of the results has size `axis1_shape`.
    """


    num_zeros = axis1_shape - events.shape[1]
    if num_zeros > 0:
        zeros = np.zeros([s if i != 1 else num_zeros for i,s in enumerate(events.shape)])
        return np.concatenate((events, zeros), axis=1)

    return events
    


from __future__ import absolute_import, division, print_function

from keras.layers import Dense, Dropout
from keras.models import Sequential
from keras.regularizers import l2

from energyflow.archs.archbase import NNBase
from energyflow.utils import iter_or_rep

__all__ = ['DNN']

###############################################################################
# DNN
###############################################################################
class DNN(NNBase):

    """Dense Neural Network architecture."""

    # DNN(*args, **kwargs)
    def process_hps(self):
        """See [`ArchBase`](#archbase) for how to pass in hyperparameters.

        **Required DNN Hyperparameters**

        - **input_dim** : _int_
            - The number of inputs to the model.
        - **dense_sizes** : {_tuple_, _list_} of _int_
            - The number of nodes in the dense layers of the model.

        **Default DNN Hyperparameters**

        - **acts**=`'relu'` : {_tuple_, _list_} of _str_
            - Activation functions(s) for the dense layers. A single string 
            will apply the same activation to all layers. See the
            [Keras activations docs](https://keras.io/activations/) for 
            more detail.
        - **k_inits**=`'he_uniform'` : {_tuple_, _list_} of _str_
            - Kernel initializers for the dense layers. A single string 
            will apply the same initializer to all layers. See the
            [Keras initializer docs](https://keras.io/initializers/) for 
            more detail.
        - **dropouts**=`0` : {_tuple_, _list_} of _float_
            - Dropout rates for the dense layers. A single float will
            apply the same dropout rate to all layers. See the [Keras
            Dropout layer](https://keras.io/layers/core/#dropout) for more 
            detail.
        - **l2_regs**=`0` : {_tuple_, _list_} of _float_
            - $L_2$-regulatization strength for both the weights and biases
            of the dense layers. A single float will apply the same
            $L_2$-regulatization to all layers.
        """

        # process generic NN hps
        super(DNN, self).process_hps()

        # required hyperparameters
        self.input_dim = self.hps['input_dim']
        self.dense_sizes = self.hps['dense_sizes']

        # activations
        self.acts = iter_or_rep(self.hps.get('acts', 'relu'))

        # initializations
        self.k_inits = iter_or_rep(self.hps.get('k_inits', 'he_uniform'))

        # regularization
        self.dropouts = iter_or_rep(self.hps.get('dropouts', 0))
        self.l2_regs = iter_or_rep(self.hps.get('l2_regs', 0))

    def construct_model(self):

        # fresh model
        self._model = Sequential()

        # iterate over specified dense layers
        z = zip(self.dense_sizes, self.acts, self.dropouts, self.l2_regs, self.k_inits)
        for i,(dim, act, dropout, l2_reg, k_init) in enumerate(z):

            # construct variable argument dict
            kwargs = {} if i > 0 else {'input_dim': self.input_dim}
            if l2_reg > 0:
                kwargs.update({'kernel_regularizer': l2(l2_reg),
                               'bias_regularizer': l2(l2_reg)})

            # add dense layer
            self.model.add(Dense(dim, activation=act, kernel_initializer=k_init, 
                                      name='dense_'+str(i), **kwargs))

            # add dropout layer if nonzero
            if dropout > 0:
                self.model.add(Dropout(dropout, name='dropout_' + str(i)))

        # output layer
        self.model.add(Dense(self.output_dim, activation=self.output_act, name='output'))    

        # compile model
        self.compile_model()



r"""Energy Flow Networks (EFNs) and Particle Flow Networks (PFNs)
are model architectures designed for learning from collider events
as unordered, variable-length sets of particles. Both EFNs and PFNs are
parameterized by a learnable per-particle function $\Phi$ and latent space function $F$.


An EFN takes the following form:
$$\text{EFN}=F\left(\sum_{i=1}^M z_i \Phi(\hat p_i)\right)$$
where $z_i$ is a measure of the energy of particle $i$, such as $z_i = p_{T,i}$, and $\hat p_i$ is a measure 
of the angular information of particle $i$, such as $\hat p_i = (y_i,\phi_i)$.
Any infrared- and collinear-safe observable can be parameterized in this form.

A PFN takes the following form:
$$\text{PFN}=F\left(\sum_{i=1}^M \Phi(p_i)\right)$$
where $p_i$ is the information of particle $i$, such as its four-momentum, charge, or flavor.
Any observable can be parameterized in this form.
See the [Deep Sets](https://arxiv.org/abs/1703.06114) framework for additional discussion.

Since these architectures are not
used by the core EnergyFlow code, and require the external 
[Keras](https://keras.io) and [scikit-learn](http://scikit-learn.org/)
libraries, they are not imported by default but must be explicitly 
imported, e.g. `from energyflow.archs import *`.
EnergyFlow also contains several additional model architectures for ease of using
common models that frequently appear in the intersection of 
particle physics and machine learning. 
"""
from __future__ import absolute_import

import warnings

__all__ = []

# requires keras
try:
    from . import cnn
    from . import dnn
    from . import efn
    from .cnn import *
    from .dnn import *
    from .efn import *

    __all__ += cnn.__all__ + dnn.__all__ + efn.__all__

except ImportError as e:
    warnings.warn('could not import some architectures - ' + str(e))

# requires sklearn
try:
    from . import linear
    from .linear import *

    __all__ += linear.__all__

except ImportError as e:
    warnings.warn('could not import some architectures - ' + str(e))



"""Base and helper classes for EFPs."""
from __future__ import absolute_import, division, print_function

from abc import ABCMeta, abstractmethod, abstractproperty
from collections import Counter
import multiprocessing
import sys

import numpy as np
from six import with_metaclass

from energyflow.algorithms import einsum
from energyflow.measure import Measure
from energyflow.utils import create_pool, timing, transfer

###############################################################################
# EFPBase
###############################################################################
class EFPBase(with_metaclass(ABCMeta, object)):

    def __init__(self, measure, beta, kappa, normed, coords, check_input):

        if 'efpm' in measure:
            raise ValueError('\'efpm\' no longer supported')
        if 'efm' in measure:
            raise ValueError('\'efm\' no longer supported')

        # store measure object
        self._measure = Measure(measure, beta, kappa, normed, coords, check_input)

    def get_zs_thetas_dict(self, event, zs, thetas):
        if event is not None:
            zs, thetas = self._measure.evaluate(event)
        elif zs is None or thetas is None:
            raise TypeError('if event is None then zs and/or thetas cannot also be None')
        return zs, {w: thetas**w for w in self._weight_set}

    @abstractproperty
    def _weight_set(self):
        pass

    @property
    def measure(self):
        return self._measure.measure

    @property
    def beta(self):
        return self._measure.beta

    @property
    def kappa(self):
        return self._measure.kappa

    @property
    def normed(self):
        return self._measure.normed

    @property
    def coords(self):
        return self._measure.coords

    @property
    def check_input(self):
        return self._measure.check_input

    def _batch_compute_func(self, event):
        return self.compute(event, batch_call=True)

    @abstractmethod
    def compute(self, *args, **kwargs):
        pass

    def batch_compute(self, events, n_jobs=None):
        """Computes the value of the EFP on several events.

        **Arguments**

        - **events** : array_like or `fastjet.PseudoJet`
            - The events as an array of arrays of particles in coordinates
            matching those anticipated by `coords`.
        - **n_jobs** : _int_ or `None`
            - The number of worker processes to use. A value of `None` will
            use as many processes as there are CPUs on the machine.

        **Returns**

        - _1-d numpy.ndarray_
            - A vector of the EFP value for each event.
        """

        if n_jobs is None:
            n_jobs = multiprocessing.cpu_count() or 1
        chunksize = max(1, len(events)//n_jobs)

        # setup processor pool
        with create_pool(n_jobs) as pool:
            results = np.asarray(list(pool.map(self._batch_compute_func, events, chunksize)))

        return results

###############################################################################
# EFPElem
###############################################################################
class EFPElem(object):

    # if weights are given, edges are assumed to be simple 
    def __init__(self, edges, weights=None, einstr=None, einpath=None, k=None):

        transfer(self, locals(), ['einstr', 'einpath', 'k'])

        self.process_edges(edges, weights)

        self.pow2d = 2**self.d
        self.ndk = (self.n, self.d, self.k)

    def process_edges(self, edges, weights):

        # deal with arbitrary vertex labels
        vertex_set = frozenset(v for edge in edges for v in edge)
        vertices = {v: i for i,v in enumerate(vertex_set)}
        
        # determine number of vertices, empty edges are interpretted as graph with one vertex
        self.n = len(vertices) if len(vertices) > 0 else 1

        # construct new edges with remapped vertices
        self.edges = [tuple(vertices[v] for v in sorted(edge)) for edge in edges]

        # get weights
        if weights is None:
            self.simple_edges = list(frozenset(self.edges))
            counts = Counter(self.edges)
            self.weights = tuple(counts[edge] for edge in self.simple_edges)

            # invalidate einsum quantities because edges got reordered
            self.einstr = self.einpath = None
        else:
            if len(weights) != len(self.edges):
                raise ValueError('length of weights is not number of edges')
            self.simple_edges = self.edges
            self.weights = tuple(weights)
        self.edges = [e for w,e in zip(self.weights, self.simple_edges) for i in range(w)]

        self.e = len(self.simple_edges)
        self.d = sum(self.weights)
        self.weight_set = frozenset(self.weights)

    def compute(self, zs, thetas_dict):
        einsum_args = [thetas_dict[w] for w in self.weights] + self.n*[zs]
        return einsum(self.einstr, *einsum_args, optimize=self.einpath)

    def set_timer(self):
        self.times = []
        self.compute = timing(self, self.compute)

    # properties set above:
    #     n, e, d, k, ndk, edges, simple_edges, weights, weight_set, einstr, einpath,
    #     efm_einstr, efm_einpath, efm_spec, M_thresh



r"""The Energy Flow Polynomials (EFPs) are a set of observables, indexed by non-isomorphic 
multigraphs, which linearly span the space of infrared and collinear safe (IRC-safe) 
observables.

An EFP, indexed by a multigraph $G$, takes the following form:
$$\text{EFP}_G=\sum_{i_1=1}^M\cdots\sum_{i_N=1}^Mz_{i_1}\cdots z_{i_N}
\prod_{(k,\ell)\in G}\theta_{i_ki_\ell}$$
where $z_i$ is a measure of the energy of particle $i$ and $\theta_{ij}$ is a measure 
of the angular separation between particles $i$ and $j$. The specific choices for "energy"
and "angular" measure depend on the collider context and are discussed in the 
[Measures](../measures) section.
"""
from __future__ import absolute_import, division, print_function

from  itertools import chain, repeat
import re
import warnings

import numpy as np

from energyflow.algorithms import VariableElimination, einsum_path
from energyflow.efpbase import EFPBase, EFPElem
from energyflow.gen import Generator
from energyflow.utils import concat_specs, default_efp_file
from energyflow.utils.graph_utils import graph_union

__all__ = ['EFP', 'EFPSet']

###############################################################################
# EFP helpers
###############################################################################

comp_map = {
    '>':  '__gt__', 
    '<':  '__lt__', 
    '>=': '__ge__', 
    '<=': '__le__',
    '==': '__eq__', 
    '!=': '__ne__'
}

# applies comprison comp of obj on val
def explicit_comp(obj, comp, val):
    return getattr(obj, comp_map[comp])(val)

# raises TypeError if unexpected keyword left in kwargs
def kwargs_check(name, kwargs, allowed=[]):
    for k in kwargs:
        if k in allowed:
            continue
        raise TypeError(name + '() got an unexpected keyword argument \'{}\''.format(k))

_sel_re = re.compile(r'(\w+)(<|>|==|!=|<=|>=)(\d+)$')


###############################################################################
# EFP
###############################################################################
class EFP(EFPBase):

    """A class for representing and computing a single EFP. Note that all
    keyword arguments are stored as properties of the `EFP` instance.
    """

    def __init__(self, edges, measure='hadr', beta=1, kappa=1, normed=True, coords=None, 
                              check_input=True, np_optimize='greedy'):
        """
        **Arguments**

        - **edges** : _list_
            - Edges of the EFP graph specified by pairs of vertices.
        - **measure** : {`'hadr'`, `'hadrdot'`, `'ee'`}
            - The choice of measure. See [Measures](../measures) 
            for additional info.
        - **beta** : _float_
            - The parameter $\\beta$ appearing in the measure.
            Must be greater than zero.
        - **kappa** : {_float_, `'pf'`}
            - If a number, the energy weighting parameter $\\kappa$.
            If `'pf'`, use $\\kappa=v-1$ where $v$ is the valency of the vertex.
        - **normed** : _bool_
            - Controls normalization of the energies in the measure.
        - **coords** : {`'ptyphim'`, `'epxpypz'`, `None`}
            - Controls which coordinates are assumed for the input. See 
            [Measures](../measures) for additional info.
        - **check_input** : _bool_
            - Whether to check the type of the input each time or assume
            the first input type.
        - **np_optimize** : {`True`, `False`, `'greedy'`, `'optimal'`}
            - The `optimize` keyword of `numpy.einsum_path`.
        """

        # initialize EFPBase
        super(EFP, self).__init__(measure, beta, kappa, normed, coords, check_input)

        # store these edges as an EFPElem
        self.efpelem = EFPElem(edges)
        self._np_optimize = np_optimize

        # setup ve for standard efp compute
        (self.efpelem.einstr, 
         self.efpelem.einpath, 
         self.chi) = VariableElimination(np_optimize).einspecs(self.simple_graph, self.n)

            
    #===============
    # public methods
    #===============

    # compute(event=None, zs=None, thetas=None)
    def compute(self, event=None, zs=None, thetas=None, batch_call=None):
        """Computes the value of the EFP on a single event.

        **Arguments**

        - **event** : 2-d array_like or `fastjet.PseudoJet`
            - The event as an array of particles in the coordinates specified
            by `coords`.
        - **zs** : 1-d array_like
            - If present, `thetas` must also be present, and `zs` is used in place 
            of the energies of an event.
        - **thetas** : 2-d array_like
            - If present, `zs` must also be present, and `thetas` is used in place 
            of the pairwise angles of an event.

        **Returns**

        - _float_
            - The EFP value.
        """

        zs, thetas_dict = self.get_zs_thetas_dict(event, zs, thetas)
        return self.efpelem.compute(zs, thetas_dict)

    #===========
    # properties
    #===========

    @property
    def _weight_set(self):
        """Set of edge weights for the graph of this EFP."""

        return self.efpelem.weight_set

    @property
    def _einstr(self):
        """Einstein summation string for the EFP computation."""

        return self.efpelem.einstr

    @property
    def _einpath(self):
        """Numpy einsum path specification for EFP computation."""

        return self.efpelem.einpath

    @property
    def np_optimize(self):
        """The np_optimize keyword argument that initialized this EFP instance."""

        return self._np_optimize

    @property
    def graph(self):
        """Graph of this EFP represented by a list of edges."""

        return self.efpelem.edges

    @property
    def simple_graph(self):
        """Simple graph of this EFP (forgetting all multiedges)
        represented by a list of edges."""

        return self.efpelem.simple_edges

    @property
    def n(self):
        """Number of vertices in the graph of this EFP."""

        return self.efpelem.n

    @property
    def d(self):
        """Degree, or number of edges, in the graph of this EFP."""

        return self.efpelem.d

    @property
    def e(self):
        """Number of edges in the simple graph of this EFP."""

        return self.efpelem.e

    @property
    def c(self):
        """VE complexity $\\chi$ of this EFP."""

        return self.chi


###############################################################################
# EFPSet
###############################################################################
class EFPSet(EFPBase):

    """A class that holds a collection of EFPs and computes their values on events.
    Note that all keyword arguments are stored as properties of the `EFPSet` instance.
    """

    # EFPSet(*args, filename=None, measure='hadr', beta=1, kappa=1, normed=True, 
    #        coords='ptyphim', check_input=True, verbose=False)
    def __init__(self, *args, **kwargs):
        r"""EFPSet can be initialized in one of three ways (in order of precedence):

        1. **Default** - Use the ($d\le10$) EFPs that come installed with the
        `EnergFlow` package.
        2. **Generator** - Pass in a custom `Generator` object as the
        first positional argument.
        3. **Custom File** - Pass in the name of a `.npz` file saved
        with a custom `Generator`.

        To control which EFPs are included, `EFPSet` accepts an arbitrary
        number of specifications (see [`sel`](#sel)) and only EFPs meeting each
        specification are included in the set.

        **Arguments**

        - ***args** : _arbitrary positional arguments_
            - If the first positional argument is a `Generator` instance,
            it is used for initialization. The remaining positional
            arguments must be valid arguments to `sel`.
        - **filename** : _string_
            - Path to a `.npz` file which has been saved by a valid
            `energyflow.Generator`.
        - **measure** : {`'hadr'`, `'hadr-dot'`, `'ee'`}
            - See [Measures](../measures) for additional info.
        - **beta** : _float_
            - The parameter $\\beta$ appearing in the measure.
            Must be greater than zero.
        - **kappa** : {_float_, `'pf'`}
            - If a number, the energy weighting parameter $\\kappa$.
            If `'pf'`, use $\\kappa=v-1$ where $v$ is the valency of the vertex.
        - **normed** : _bool_
            - Controls normalization of the energies in the measure.
        - **coords** : {`'ptyphim'`, `'epxpypz'`, `None`}
            - Controls which coordinates are assumed for the input. See 
            [Measures](../measures) for additional info.
        - **check_input** : _bool_
            - Whether to check the type of the input each time or assume
            the first input type.
        - **verbose** : _bool_
            - Controls printed output when initializing EFPSet.
        """

        default_kwargs = {'filename': None,
                          'measure': 'hadr',
                          'beta': 1,
                          'kappa': 1,
                          'normed': True,
                          'coords': None,
                          'check_input': True,
                          'verbose': False}
        measure_kwargs = ['measure', 'beta', 'kappa', 'normed', 'coords', 'check_input']

        # process arguments
        for k,v in default_kwargs.items():
            if k not in kwargs:
                kwargs[k] = v
            if k not in measure_kwargs:
                setattr(self, k, kwargs.pop(k))

        kwargs_check('__init__', kwargs, allowed=measure_kwargs)

        # initialize EFPBase
        super(EFPSet, self).__init__(*[kwargs[k] for k in measure_kwargs])

        # handle different methods of initialization
        maxs = ['nmax', 'emax', 'dmax', 'cmax', 'vmax', 'comp_dmaxs']
        elemvs = ['edges', 'weights', 'einstrs', 'einpaths']
        if len(args) >= 1 and isinstance(args[0], Generator):
            constructor_attrs = maxs + elemvs + ['cols', 'c_specs', 'disc_specs', 'disc_formulae']
            gen = {attr: getattr(args[0], attr) for attr in constructor_attrs}
            args = args[1:]
        elif self.filename is not None:
            self.filename += '.npz' if not self.filename.endswith('.npz') else ''
            gen = np.load(self.filename)
        else:
            gen = np.load(default_efp_file)

        # compile regular expression for use in sel()
        self._sel_re = _sel_re
        
        # put column headers and indices into namespace
        self._cols = gen['cols']
        self._set_col_inds()

        # put gen maxs into dict
        self.gen_maxs = {m: gen[m] for m in maxs}

        # get disc formulae and disc mask
        orig_disc_specs = gen['disc_specs']
        disc_mask = self.sel(*args, specs=orig_disc_specs)
        self.disc_formulae = gen['disc_formulae'][disc_mask]

        # get connected specs and full specs
        orig_c_specs = gen['c_specs']
        c_mask = self.sel(*args, specs=orig_c_specs)
        self._cspecs = orig_c_specs[c_mask]
        self._specs = concat_specs(self._cspecs, orig_disc_specs[disc_mask])

        # make EFPElem list
        z = zip(*([gen[v] for v in elemvs] + [orig_c_specs[:,self.k_ind]]))
        self.efpelems = [EFPElem(*args) for m,args in enumerate(z) if c_mask[m]]

        # union over all weights needed
        self.__weight_set = frozenset(w for efpelem in self.efpelems for w in efpelem.weight_set)

        # get col indices for disconnected formulae
        connected_ndk = {efpelem.ndk: i for i,efpelem in enumerate(self.efpelems)}
        self.disc_col_inds = []
        for formula in self.disc_formulae:
            try:
                self.disc_col_inds.append([connected_ndk[factor] for factor in formula])
            except KeyError:
                warnings.warn('connected efp needed for {} not found'.format(formula))

        # handle printing
        if self.verbose:
            print('Originally Available EFPs:')
            self.print_stats(specs=concat_specs(orig_c_specs, orig_disc_specs), lws=2)
            if len(args) > 0:
                print('Current Stored EFPs:')
                self.print_stats(lws=2)

    #================
    # PRIVATE METHODS
    #================

    def _set_col_inds(self):
        self.__dict__.update({col+'_ind': i for i,col in enumerate(self.cols)})

    def _make_graphs(self, connected_graphs):
        disc_comps = [[connected_graphs[i] for i in col_inds] for col_inds in self.disc_col_inds]
        return np.asarray(connected_graphs + [graph_union(*dc) for dc in disc_comps])

    #===============
    # PUBLIC METHODS
    #===============

    # calc_disc(X)
    def calc_disc(self, X):
        """Computes disconnected EFPs according to the internal 
        specifications using the connected EFPs provided as input.

        **Arguments**

        - **X** : _numpy.ndarray_
            - Array of connected EFPs. Rows are different events, columns 
            are the different EFPs. Can handle a single event (a 1-dim array) 
            as input. EFPs are assumed to be in the order expected by the 
            instance of `EFPSet`; the safest way to ensure this is to use 
            the same `EFPSet` to calculate both connected and disconnected 
            EFPs. This function is used internally in `compute` and 
            `batch_compute`.

        **Returns**

        - _numpy.ndarray_
            - A concatenated array of the connected and disconnected EFPs.
        """

        if len(self.disc_col_inds) == 0:
            return np.asarray(X)

        X = np.atleast_2d(X)

        results = np.empty((len(X), len(self.disc_col_inds)), dtype=float)
        for i,formula in enumerate(self.disc_col_inds):
            results[:,i] = np.prod(X[:,formula], axis=1)

        return np.squeeze(np.concatenate((X, results), axis=1))

    # compute(event=None, zs=None, thetas=None)
    def compute(self, event=None, zs=None, thetas=None, batch_call=False):
        """Computes the values of the stored EFPs on a single event.

        **Arguments**

        - **event** : 2-d array_like or `fastjet.PseudoJet`
            - The event as an array of particles in the coordinates specified
            by `coords`.
        - **zs** : 1-d array_like
            - If present, `thetas` must also be present, and `zs` is used in place 
            of the energies of an event.
        - **thetas** : 2-d array_like
            - If present, `zs` must also be present, and `thetas` is used in place 
            of the pairwise angles of an event.

        **Returns**

        - _1-d numpy.ndarray_
            - A vector of the EFP values.
        """
        
        zs, thetas_dict = self.get_zs_thetas_dict(event, zs, thetas)
        results = [efpelem.compute(zs, thetas_dict) for efpelem in self.efpelems]

        if batch_call:
            return results
        else:
            return self.calc_disc(results)

    def batch_compute(self, events, n_jobs=None):
        """Computes the value of the stored EFPs on several events.

        **Arguments**

        - **events** : array_like or `fastjet.PseudoJet`
            - The events as an array of arrays of particles in coordinates
            matching those anticipated by `coords`.
        - **n_jobs** : _int_ or `None`
            - The number of worker processes to use. A value of `None` will attempt
            to use as many processes as there are CPUs on the machine.

        **Returns**

        - _2-d numpy.ndarray_
            - An array of the EFP values for each event.
        """

        return self.calc_disc(super(EFPSet, self).batch_compute(events, n_jobs))

    # sel(*args)
    def sel(self, *args, **kwargs):
        """Computes a boolean mask of EFPs matching each of the
        specifications provided by the `args`. 

        **Arguments**

        - ***args** : arbitrary positional arguments
            - Each argument can be either a string or a length-two 
            iterable. If the argument is a string, it should consist 
            of three parts: a character which is a valid element of 
            `cols`, a comparison operator (one of `<`, `>`, `<=`, 
            `>=`, `==`, `!=`), and a number. Whitespace between the 
            parts does not matter. If the argument is a tuple, the 
            first element should be a string containing a column 
            header character and a comparison operator; the second 
            element is the value to be compared. The tuple version 
            is useful when the value is a variable that changes 
            (such as in a list comprehension).

        **Returns**

        - _1-d numpy.ndarray_
            - A boolean array of length the number of EFPs stored by this object. 
        """

        # ensure only valid keyword args are passed
        specs = kwargs.pop('specs', None)
        kwargs_check('sel', kwargs)

        # use default specs if non provided
        if specs is None:
            specs = self.specs

        # iterate through arguments
        mask = np.ones(len(specs), dtype=bool)
        for arg in args:

            # parse arg
            if isinstance(arg, str):
                s = arg.replace(' ', '')
            elif hasattr(arg, '__getitem__'):
                if len(arg) == 2:
                    s = arg[0].replace(' ', '') + str(arg[1])
                else:
                    raise ValueError('{} is not length 2'.format(arg))
            else:
                raise TypeError('invalid type for {}'.format(arg))

            # match string to pattern
            match = self._sel_re.match(s)
            if match is None:
                raise ValueError('could not understand \'{}\''.format(arg))

            # get the variable of the selection
            var = match.group(1)
            if var not in self.cols:
                raise ValueError('\'{}\' not in {}'.format(var, self.cols))

            # get the comparison and value
            comp, val = match.group(2, 3)

            # AND the selection with mask
            mask &= explicit_comp(specs[:,getattr(self, var+'_ind')], comp, int(val))
            
        return mask

    # csel(*args)
    def csel(self, *args):
        """Same as `sel` except using `cspecs` to select from."""

        return self.sel(*args, specs=self.cspecs)

    # count(*args)
    def count(self, *args, **kwargs):
        """Counts the number of EFPs meeting the specifications
        of the arguments using `sel`.

        **Arguments** 

        - ***args** : arbitrary positional arguments
            - Valid arguments to be passed to `sel`.

        **Returns**

        - _int_
            - The number of EFPs meeting the specifications provided.
        """

        return np.count_nonzero(self.sel(*args, **kwargs))

    # graphs(*args)
    def graphs(self, *args):
        """Graphs meeting provided specifications.

        **Arguments** 

        - ***args** : arbitrary positional arguments
            - Valid arguments to be passed to `sel`, or, if a single integer, 
            the index of a particular graph.

        **Returns**

        - _list_, if single integer argument is given
            - The list of edges corresponding to the specified graph
        - _1-d numpy.ndarray_, otherwise
            - An array of graphs (as lists of edges) matching the specifications.
        """

        # if we haven't extracted the graphs, do it now
        if not hasattr(self, '_graphs'):
            self._graphs = self._make_graphs([elem.edges for elem in self.efpelems])

        # handle case of single graph
        if len(args) and isinstance(args[0], int):
            return self._graphs[args[0]]

        # filter graphs based on mask
        return self._graphs[self.sel(*args)]

    # simple_graphs(*args)
    def simple_graphs(self, *args):
        """Simple graphs meeting provided specifications.

        **Arguments** 

        - ***args** : arbitrary positional arguments
            - Valid arguments to be passed to `sel`, or, if a single integer, 
            the index of particular simple graph.

        **Returns**

        - _list_, if single integer argument is given
            - The list of edges corresponding to the specified simple graph
        - _1-d numpy.ndarray_, otherwise
            - An array of simple graphs (as lists of edges) matching the specifications.
        """

        # is we haven't extracted the simple graphs, do it now
        if not hasattr(self, '_simple_graphs'):
            self._simple_graphs = self._make_graphs([elem.simple_edges for elem in self.efpelems])

        # handle case of single graph
        if len(args) and isinstance(args[0], int):
            return self._simple_graphs[args[0]]

        # filter simple graphs based on mask
        return self._simple_graphs[self.sel(*args)]

    def print_stats(self, specs=None, lws=0):
        if specs is None:
            specs = self.specs
        num_prime = self.count('p==1', specs=specs)
        num_composite = self.count('p>1', specs=specs)
        pad = ' '*lws
        print(pad + 'Prime:', num_prime)
        print(pad + 'Composite:', num_composite)
        print(pad + 'Total: ', num_prime+num_composite)

    def set_timers(self):
        for efpelem in self.efpelems:
            efpelem.set_timer()

    def get_times(self):
        efp_times = np.asarray([elem.times for elem in self.efpelems])
        return efp_times

    #===========
    # properties
    #===========

    @property
    def _weight_set(self):
        return self.__weight_set

    @property
    def cols(self):
        """Column labels for `specs`. 
        Those of primary interest are listed below.

        - `n` : Number of vertices.
        - `e` : Number of simple edges.
        - `d` : Degree, or number of multiedges.
        - `v` : Maximum valency (number of edges touching a vertex).
        - `k` : Unique identifier within EFPs of this (n,d).
        - `c` : VE complexity $\\chi$.
        - `p` : Number of prime factors (or connected components).
        - `h` : Number of valency 1 vertices (a.k.a. 'hanging chads').
        """

        return self._cols

    @property
    def specs(self):
        """An array of EFP specifications. Each row represents an EFP 
        and the columns represent the quantities indicated by `cols`."""

        return self._specs

    @property
    def cspecs(self):
        """Specification array for connected EFPs."""

        return self._cspecs



"""The EnergyFlow package."""
from __future__ import absolute_import

# import toplevel submodules
from . import algorithms
from . import efp
from . import efpbase
from . import gen
from . import measure
from . import utils

# import toplevel attributes
from .efp import *
from .gen import *
from .measure import *
from .utils import *

__all__ = (gen.__all__ + 
           efp.__all__ + 
           measure.__all__ + 
           utils.__all__)

__version__ = '0.11.0'



r"""### Particle Tools

Tools to compute particle kinematic quantities from four-vectors,
such as transverse momentum $p_T$, rapidity $y$, and azimuthal angle
$\phi$, and vice versa.
"""
from __future__ import absolute_import, division

import warnings

import numpy as np

__all__ = [
    'p4s_from_ptyphims',
    'p4s_from_ptyphis',
    'ptyphims_from_p4s',
    'pts_from_p4s',
    'ys_from_p4s',
    'phis_from_p4s',
    'ms_from_p4s',
    'phi_fix',
    'flat_metric',
]

def ptyphims_from_p4s(p4s, phi_ref=None, keep_allzeros=True):
    r"""Compute the `[pt,y,phi,m]` representation of a four-vector for each Euclidean
    four-vector given as input. All-zero four-vectors are removed unless `keep_shape` 
    is `True`.

    **Arguments**

    - **p4s** : _numpy.ndarray_ or _list_
        - An event as an `(M,4)` array of four-vectors `[E,px,py,pz]` for each particle.
        A single particle as a one-dimensional array or list is also accepted.
    - **phi_ref** : _float_
        - A reference value used so that all phis will be within $\pm\pi$ of this value.
        A value of `None` means that no phi fixing will be applied.
    - **keep_allzeros** : _bool_
        - Flag to determine if all-zero four-vectors will be retained as such. Otherwise,
        they are removed (resulting in a change in the shape of the output).

    **Returns**

    - _numpy.ndarray_
        - An array of size `(M,4)` consisting of the transverse momentum, rapidity,
        azimuthal angle, and mass of each particle. If a single particle was given as 
        input, a one-dimensional array is returned.
    """

    # ensure a two-dimensional array
    particles = np.copy(np.atleast_2d(p4s))

    # find non-zero particles
    nonzero_mask = np.count_nonzero(particles, axis=1) > 0
    nonzero_particles = particles[nonzero_mask]

    # get quantities
    pts = pts_from_p4s(nonzero_particles)
    ys = ys_from_p4s(nonzero_particles)
    phis = phis_from_p4s(nonzero_particles, phi_ref=phi_ref)
    ms = ms_from_p4s(nonzero_particles)
    ptyphims = np.vstack((pts, ys, phis, ms)).T

    # keep the all-zero particles
    if keep_allzeros:
        particles[nonzero_mask] = ptyphims
        return np.squeeze(particles)

    # return just the ptyphims for the non-zero particles
    return np.squeeze(ptyphims)


def pts_from_p4s(p4s):
    """Calculate the transverse momenta of a collection of four-vectors
    
    **Arguments**

    - **p4s** : _numpy.ndarray_ or _list_
        - An event as an `(M,4)` array of four-vectors `[E,px,py,pz]` for each particle.
        A single particle as a one-dimensional array or list is also accepted.

    **Returns**

    - _numpy.ndarray_
        - An `M`-length array consisting of the transverse momentum of each particle.
        If a single particle was given as input, a single float is returned.
    """

    pts = np.sqrt(p4s[...,1]**2 + p4s[...,2]**2)
    return np.squeeze(pts)


def ys_from_p4s(p4s):
    """Calculate the rapidities of a collection of four-vectors
    
    **Arguments**

    - **p4s** : _numpy.ndarray_ or _list_
        - An event as an `(M,4)` array of four-vectors `[E,px,py,pz]` for each particle.
        A single particle as a one-dimensional array or list is also accepted.

    **Returns**

    - _numpy.ndarray_
        - An `M`-length array consisting of the rapidity of each particle.
        If a single particle was given as input, a single float is returned.
    """

    ys = 0.5*np.log((p4s[...,0] + p4s[...,3])/(p4s[...,0] - p4s[...,3]))
    return np.squeeze(ys)


def phis_from_p4s(p4s, phi_ref=None):
    r"""Calculate the azimuthal angles of a collection of four-vectors. If `phi_ref` is 
    not `None`, then `phi_fix` is called using this value. Otherwise, 
    the angles are chosen to be in the inverval $[0,2\pi]$.
    
    **Arguments**

    - **p4s** : _numpy.ndarray_ or _list_
        - An event as an `(M,4)` array of four-vectors `[E,px,py,pz]` for each particle.
        A single particle as a one-dimensional array or list is also accepted.
    - **phi_ref** : _float_
        - See [`phi_fix`](#phi_fix)

    **Returns**

    - _numpy.ndarray_
        - An `M`-length array consisting of the azimuthal angle of each particle.
        If a single particle was given as input, a single float is returned.
    """

    phis = np.arctan2(p4s[...,2], p4s[...,1])
    phis[phis<0] += 2*np.pi

    # ensure close to reference value
    if phi_ref is not None:
        phis = phi_fix(phis, phi_ref)

    return np.squeeze(phis)


def ms_from_p4s(p4s):
    """Calculate the masses of a collection of four-vectors.
    
    **Arguments**

    - **p4s** : _numpy.ndarray_ or _list_
        - An event as an `(M,4)` array of four-vectors `[E,px,py,pz]` for each particle.
        A single particle as a one-dimensional array or list is also accepted.

    **Returns**

    - _numpy.ndarray_
        - An `M`-length array consisting of the mass of each particle. If a single 
        particle was given as input, a single float is returned.
    """

    p4s = np.atleast_2d(p4s)
    m2s = np.squeeze(p4s[...,0]**2 - np.sum(p4s[...,1:]**2, axis=-1))
    ms = np.sign(m2s)*np.sqrt(np.abs(m2s))
    return np.squeeze(ms)


def p4s_from_ptyphims(ptyphims):
    """Calculate Euclidean four-vectors from transverse momentum, rapidity, azimuthal angle,
    and (optionally) mass for each input.
    
    **Arguments**

    - **ptyphims** : _numpy.ndarray_ or _list_
        - An array with shape `(M,4)` of `[pT,y,phi,m]` for each particle. An array with 
        shape `(M,3)` is also accepted where the masses are taken to be zero. A single 
        particle is also accepted.

    **Returns**

    - _numpy.ndarray_
        - An event as an `(M,4)` array of four-vectors `[E,px,py,pz]` for each particle.
        If a single particle was given as input, a single four-vector will be returned.
    """

    # ensure a two-dimensional array
    ptyphims = np.atleast_2d(ptyphims)

    # get pts, ys, phis
    pts, ys, phis = [ptyphims[:,i] for i in range(3)]

    # get masses
    ms = ptyphims[:,3] if ptyphims.shape[1] == 4 else np.zeros(len(ptyphims))

    Ets = np.sqrt(pts**2 + ms**2)
    p4s = np.vstack([Ets*np.cosh(ys), pts*np.cos(phis), pts*np.sin(phis), Ets*np.sinh(ys)]).T

    return np.squeeze(p4s)


def p4s_from_ptyphis(ptyphis):
    """_Legacy function_: Will be removed in version 1.0. Use `p4s_from_ptyphims`
    for equivalent functionality.
    """

    warnings.warn('This function is deprecated and will be removed in version 1.0. ' 
                  'Use p4s_from_ptyphims for equivalent functionality.')

    return p4s_from_ptyphims(ptyphis)


twopi = 2*np.pi
def phi_fix(phis, phi_ref, copy=False):
    r"""A function to ensure that all phi values are within $\pi$ of `phi_ref`. 
    It is assumed that all starting phi values are within $2\pi$ of `phi_ref`.

    **Arguments**

    - **phis** : _numpy.ndarray_ or _list_
        - One-dimensional array of phi values.
    - **phi_ref** : _float_
        - A reference value used so that all phis will be within $\pm\pi$ of this value.
    - **copy** : _bool_
        - Determines if `phis` are copied or not. If `False` then `phis` may be 
        modified in place.

    **Returns**

    - _numpy.ndarray_
        - An array of the fixed phi values.
    """

    phis = np.asarray(phis)
    diff = phis - phi_ref
    new_phis = np.copy(phis) if copy else phis
    new_phis[diff > np.pi] -= twopi
    new_phis[diff < -np.pi] += twopi
    return new_phis

long_metric = np.array([1.] + [-1.]*100)
def flat_metric(dim):
    """The Minkowski metric in `dim` spacetime dimensions in the mostly-minus convention.
    
    **Arguments**

    - **dim** : _int_
        - The number of spacetime dimensions (thought to be four in our universe).

    **Returns**

    - _1-d numpy.ndarray_
        - A `dim`-length, one-dimensional (not matrix) array equal to `[+1,-1,...,-1]`
    """

    if dim <= 101:
        return long_metric[:dim]
    return np.asarray([1.] + [-1.]*(dim-1))
